from dataclasses import dataclass

import cv2
import numpy as np


@dataclass
class ConnectedComponent:
    """Dataclass to represent a connected component in a binary mask.

    A connected component is a set of pixels that are connected together.
    The bounding box is the smallest rectangle that contains all the pixels of the connected component.

    Attributes:
        mask: The binary mask of the connected component.
        bounding_box: The bounding box of the connected component. It is represented as a numpy array
        with the format [left, top, width, height].
    """

    mask: np.ndarray
    bounding_box: np.ndarray
    area: int


def extract_connected_components_from_binary_mask(
    mask: np.ndarray, min_area: int = 1
) -> list[ConnectedComponent]:
    """Find the connected components and their corresponding bounding boxes in the binary masks.

    Args:
        mask: binary mask with shape (H, W, 1).
        min_area: The minimum area of a connected component to be considered valid.

    Returns:
        An list of ConnectedComponent corresponding to the valid connected component.
    """
    if (mask == 0).all():
        return []

    num_labels, labels, stats, _ = cv2.connectedComponentsWithStats(
        mask.astype(np.uint8)
    )

    valid_cc = []
    for label_idx in range(1, num_labels):  # Exclude background label (0)
        cc_stats = stats[label_idx]
        if cc_stats[cv2.CC_STAT_AREA] >= min_area:
            valid_cc.append(
                ConnectedComponent(
                    mask=(labels == label_idx).astype(np.uint8),
                    bounding_box=np.array(
                        (
                            cc_stats[cv2.CC_STAT_LEFT],
                            cc_stats[cv2.CC_STAT_TOP],
                            cc_stats[cv2.CC_STAT_WIDTH],
                            cc_stats[cv2.CC_STAT_HEIGHT],
                        ),
                        dtype=np.uint32,
                    ),
                    area=cc_stats[cv2.CC_STAT_AREA].item(),
                )
            )

    return valid_cc



from pathlib import Path
from typing import Callable

from PIL import Image
import torch
from torchvision.tv_tensors import Image as TvImage, Mask as TvMask

from any_gold.utils.dataset import (
    AnyVisionSegmentationDataset,
    AnyVisionSegmentationOutput,
)
from any_gold.utils.synapse import SynapseZipBase


class KPITask1PatchLevelOutput(AnyVisionSegmentationOutput):
    """Output class for KPI Task 1 Patch Level dataset.

    The label is the name of the disease.
    """

    pass


class KPITask1PatchLevel(AnyVisionSegmentationDataset, SynapseZipBase):
    """KPI Task 1 Patch Level Dataset from Synapse.

    The KPI Task 1 Patch Level dataset is introduced in
    [KPIs 2024 Challenge: Advancing Glomerular Segmentation from Patch-to Slide-Level](https://arxiv.org/pdf/2502.07288)

    The dataset is a collection of images and their corresponding segmentation masks
    for glomeruli identification in various Chronic kidney diseases (CKD).

    The dataset is downloaded from [Synapse](https://www.synapse.org/Synapse:syn63688309) and its data will be downloaded
    and stored in the specified root directory.

    There are 3 different splits available: train, val, and test.

    Attributes:
        root: The root directory where the dataset is stored.
        split: The split of the dataset to use. Can be 'train', 'val', or 'test'. Default is 'train'.
        entity: The entity ID of the dataset on Synapse.
        transform: A transform to apply to the images.
        target_transform: A transform to apply to the masks.
        transforms: A transform to apply to both images and masks.
        It cannot be set together with transform and target_transform.
        override: If True, will override the existing dataset in the root directory. Default is False.
        samples: A list of tuples containing the image path and class name.
    """

    _ENTITIES: dict[str, dict[str, str]] = {
        "train": {
            "entity": "syn60249790",
            "name": "train",
        },
        "val": {
            "entity": "syn60249847",
            "name": "validation",
        },
        "test": {
            "entity": "syn63688309",
            "name": "test",
        },
    }

    def __init__(
        self,
        root: str | Path,
        split: str = "train",
        transform: Callable | None = None,
        target_transform: Callable | None = None,
        transforms: Callable | None = None,
        override: bool = False,
    ) -> None:
        AnyVisionSegmentationDataset.__init__(
            self,
            root=root,
            transform=transform,
            target_transform=target_transform,
            transforms=transforms,
            override=override,
        )

        if split not in self._ENTITIES:
            raise ValueError(
                f"Split {split} is not available. Available splits are {self._ENTITIES.keys()}."
            )
        self.split = split
        self.entity = self._ENTITIES[split]["entity"]
        SynapseZipBase.__init__(
            self,
            root=root,
            entity=self.entity,
            override=override,
        )

    def _setup(self) -> None:
        root = self.root / self._ENTITIES[self.split]["name"]
        if self.override or not root.exists():
            self.download()

        self.samples: list[tuple[Path, str]] = [
            (image_path, class_dir.name)
            for class_dir in root.iterdir()
            for patch_dir in class_dir.iterdir()
            for image_path in (patch_dir / "img").glob("*.jpg")
        ]

    def __len__(self) -> int:
        return len(self.samples)

    def get_image_path(self, index: int) -> Path:
        """Get the path of an image."""
        return self.samples[index][0]

    def get_raw(self, index: int) -> KPITask1PatchLevelOutput:
        image_path, disease = self.samples[index]
        mask_path = image_path.parent.parent / f"mask/{image_path.stem[:-3]}mask.jpg"

        image = TvImage(Image.open(image_path).convert("RGB"), dtype=torch.uint8)
        mask = (
            TvMask(Image.open(mask_path).convert("L"), dtype=torch.uint8)
            if mask_path is not None
            else torch.zeros((1, *image.shape[-2:]), dtype=torch.uint8)
        )

        return KPITask1PatchLevelOutput(
            image=image, mask=mask, label=disease, index=index
        )



from typing import Any

import numpy as np


def vision_segmentation_stats(
    shape_counts: dict[tuple[int, ...], int],
    areas: list[int],
    object_counts: list[int],
) -> dict[str, Any]:
    """Compute statistics for vision segmentation labels.

    Args:
        shape_counts: A dictionary where keys are shapes (tuples of dimensions) and values are counters of occurrences.
        areas: A list of areas for each connected component.
        object_counts: A list of object counts in each image.

    Returns:
        A dictionary containing:
        - 'shapes': A dictionary with the shape as key and the corresponding number of images in the dataset.
        - 'areas': A dictionary with min, max, mean, and total area, or None if empty.
        - 'object count': A dictionary with min, max, mean, and total object count, or None if empty.

    Raises:
        ValueError: If `shape_counts` is empty or if any shape has a count of 0.
    """
    if not shape_counts:
        raise ValueError("shape_counts must not be empty")

    shapes_stats = {}
    for shape, count in shape_counts.items():
        if count == 0:
            raise ValueError(f"Shape {shape} has a count of 0, which is invalid.")
        shapes_stats[shape] = count

    return {
        "shapes": shapes_stats,
        "areas": None
        if not areas
        else {
            "min": np.min(areas).item(),
            "max": np.max(areas).item(),
            "mean": np.mean(areas).item(),
            "total": np.sum(areas).item(),
        },
        "object count": None
        if not object_counts
        else {
            "min": np.min(object_counts).item(),
            "max": np.max(object_counts).item(),
            "mean": np.mean(object_counts).item(),
            "total": np.sum(object_counts).item(),
        },
    }



import os
from functools import partial

import datasets
import huggingface_hub.utils
from datasets import load_dataset, DownloadConfig
from datasets.exceptions import DatasetNotFoundError
from datasets import Dataset as HFDataset

from any_gold.utils.dataset import AnyDataset

datasets.logging.set_verbosity_info()


class HuggingFaceDataset(AnyDataset):
    """Base class for Hugging Face datasets.

    Hugging Face is an open repository for machine learning datasets and models.
    It is accessible at https://huggingface.co/.

    This class specifies the basic way to download data from Hugging Face.

    The HUGGINGFACE_API_TOKEN environment variable must be set to access protected Hugging Face datasets.

    Attributes:
        root: The root directory where the dataset is stored.
        path: The path of the dataset on Hugging Face.
        hf_split: The dataset split to use from Hugging Face (e.g., 'train', 'test' but potentially others).
        override: If True, will override the existing dataset in the root directory. Default is False.
        samples: The Hugging Face dataset in torch format.
    """

    def __init__(
        self,
        path: str,
        hf_split: str | None = None,
        override: bool = False,
    ) -> None:
        super().__init__(
            root=huggingface_hub.constants.HUGGINGFACE_HUB_CACHE,
            override=override,
        )

        self.path = path
        self.hf_split = hf_split
        self.samples: HFDataset

        self._setup()

    def _setup(self) -> None:
        """download the data from Hugging face and initialize the dataset in torch format."""
        download_config = DownloadConfig(
            cache_dir=str(self.root), force_download=self.override
        )
        loader = partial(
            load_dataset,
            path=self.path,
            split=self.hf_split,
            download_config=download_config,
        )
        try:
            self.samples = loader().with_format("torch")
        except DatasetNotFoundError:
            HUGGINGFACE_API_TOKEN = os.environ.get("HUGGINGFACE_API_TOKEN")
            if HUGGINGFACE_API_TOKEN is None:
                raise ValueError(
                    "HUGGINGFACE_API_TOKEN environment variable must be set."
                )
            download_config.token = HUGGINGFACE_API_TOKEN
            self.samples = loader(token=HUGGINGFACE_API_TOKEN).with_format("torch")



from pathlib import Path
from tqdm import tqdm

import cv2
import numpy as np
from torch import Generator
from torch.utils.data import RandomSampler, DataLoader

import fiftyone as fo

from any_gold.tools.image.connected_component import (
    ConnectedComponent,
    extract_connected_components_from_binary_mask,
)
from any_gold.utils.dataset import (
    AnyVisionSegmentationOutput,
    AnyVisionSegmentationDataset,
)


def build_fo_detections_from_connected_components(
    connected_components: list[ConnectedComponent],
    label: str,
    index: int,
    save_dir: Path,
) -> fo.Detections:
    """Build FiftyOne detections from connected components.

    Args:
        connected_components: List of connected components.
        label: The label for the detections.
        index: The index of the sample.
        save_dir: Directory to save the masks. It is more efficient to save them locally than storing everything in RAM.
        If the data is not required to be kept, it can be a temporary directory.

    Returns:
        FiftyOne detections including all connected components.

    Raises:
        ValueError: If no connected components are provided.
    """
    if not connected_components:
        raise ValueError("No connected components provided to build detections.")

    detections = []
    for cc_idx, cc in enumerate(connected_components):
        mask_h, mask_w = cc.mask.shape[:2]
        bb_left, bb_top, bb_w, bb_h = cc.bounding_box
        cc_mask = cc.mask[
            bb_top : bb_top + bb_h, bb_left : bb_left + bb_w
        ]  # mask must only correspond to the bounding box
        mask_path = str(save_dir / f"{label}_{index}_{cc_idx}.png")
        cv2.imwrite(mask_path, cc_mask)

        detections.append(
            fo.Detection(
                label=label,
                bounding_box=cc.bounding_box
                / np.array(
                    [mask_w, mask_h, mask_w, mask_h]
                ),  # fiftyone expects bounding box normalized to [0, 1] from width and height values
                mask_path=mask_path,
            )
        )

    return fo.Detections(detections=detections)


def build_fo_sample_from_any_vision_segmentation_output(
    sample: AnyVisionSegmentationOutput,
    save_dir: Path,
) -> fo.Sample:
    """Build a FiftyOne sample from an AnyVisionSegmentationOutput.

    So far, it is assumed that the mask provided by an AnyVisionSegmentationOutput is a binary masks.

    Args:
        sample: The AnyVisionSegmentationOutput to convert.
        save_dir: The directory where the images and masks will be saved. It is more efficient to save them locally
        than storing everything in RAM. If the data is not required to be kept, it can be a temporary directory.

    Returns:
        A FiftyOne sample containing the image, mask, and metadata.
    """
    label = sample["label"]
    index = sample["index"]

    # save the image locally
    image = sample["image"].permute(1, 2, 0).numpy()
    image_path = save_dir / f"image_{index}.png"
    cv2.imwrite(str(image_path), image)

    # save the masks locally and create fiftyone detections
    cc_masks_and_boxes = extract_connected_components_from_binary_mask(
        sample["mask"].permute(1, 2, 0).numpy()
    )
    detections = (
        build_fo_detections_from_connected_components(
            cc_masks_and_boxes,
            label=label,
            index=index,
            save_dir=save_dir,
        )
        if cc_masks_and_boxes
        else None
    )

    return fo.Sample(
        filepath=str(image_path),
        ground_truth=detections,
        metadata=fo.ImageMetadata(
            width=image.shape[1],
            height=image.shape[0],
            num_channels=image.shape[2] if len(image.shape) == 3 else 1,
            index=index,
        ),
    )


def build_fo_dataset_from_any_vision_segmentation_dataset(
    dataset: AnyVisionSegmentationDataset,
    dataset_name: str,
    save_dir: Path,
    num_samples: int | None = None,
    persistent: bool = False,
    overwrite: bool = True,
    batch_size: int = 8,
    num_workers: int = 2,
    seed: int | None = None,
) -> fo.Dataset:
    """Build a FiftyOne dataset from an AnyVisionSegmentationDataset.

    Args:
        dataset: The AnyVisionSegmentationDataset to convert. It is assumed that the segmentation masks are binary masks.
        dataset_name: The name of the FiftyOne dataset.
        save_dir: The directory where the images and masks will be saved.
        num_samples: The number of samples to include in the dataset. If None, all samples will be included.
        persistent: If True, the dataset will be persistent in fiftyone and it will not be possible to create
        a dataset with same name without setting `overwrite` to True.
        overwrite: If True, the existing dataset will be overwritten.
        batch_size: The batch size for the DataLoader.
        num_workers: The number of workers for the DataLoader.
        seed: The random seed for sampling the dataset. If None, no specific seed is set.

    Returns:
        A FiftyOne dataset containing the samples from the AnyVisionSegmentationDataset.
    """
    fo_dataset = fo.Dataset(dataset_name, persistent=persistent, overwrite=overwrite)

    random_generator = Generator().manual_seed(seed) if seed is not None else None
    sampler = (
        RandomSampler(
            dataset,
            replacement=False,
            num_samples=num_samples,
            generator=random_generator,
        )
        if num_samples is not None
        else None
    )
    dataloader: DataLoader = DataLoader(
        dataset,
        batch_size=batch_size,
        sampler=sampler,
        num_workers=num_workers,
        collate_fn=lambda x: x,  # get directly access to the samples without collating
    )

    for batch in tqdm(dataloader, desc="Converting samples to FiftyOne format"):
        for sample in batch:
            fo_dataset.add_sample(
                build_fo_sample_from_any_vision_segmentation_output(
                    sample=sample,
                    save_dir=save_dir,
                )
            )

    return fo_dataset



import os
from abc import abstractmethod
from pathlib import Path
from tempfile import TemporaryDirectory
from logging import getLogger

from torchvision.datasets.utils import extract_archive

from zenodo_client import Zenodo

from any_gold.utils.dataset import AnyDataset

logger = getLogger(__name__)


class ZenodoDataset(AnyDataset):
    """Base class for Zenodo datasets.

    Zenodo is an open repository for research data and software.
    It is accessible at https://zenodo.org/.

    This class specifies the basic way to download data from Zenodo. Each inherited class must implement the
    `_move_data_to_root` and `_setup` methods to download the data from Zenodo.

    The ZENODO_API_TOKEN environment variable must be set to access Zenodo datasets.

    Attributes:
        root: The root directory where the dataset is stored.
        record_id: The record ID of the dataset on Zenodo.
        name: The name of the dataset on Zenodo.
        override: If True, will override the existing dataset in the root directory. Default is False.
    """

    def __init__(
        self,
        root: str | Path,
        record_id: str,
        name: str,
        override: bool = False,
    ) -> None:
        super().__init__(root=root, override=override)

        self.record_id = record_id
        self.name = name

        self._setup()

    @abstractmethod
    def _setup(self) -> None:
        """download the data from Zenodo and initialize the elements of the dataset."""

    @abstractmethod
    def _move_data_to_root(self, file: Path) -> None:
        """Make the data available in the root directory.

        This method can be used to extract the data from an archive or to reorganise the data after downloading it.
        """

    def download(self) -> None:
        """Download the data from Zenodo and store the dataset in the root folder."""
        ZENODO_API_TOKEN = os.environ.get("ZENODO_API_TOKEN")
        if ZENODO_API_TOKEN is None:
            raise ValueError(
                "Please set the ZENODO_API_TOKEN environment variable to access Zenodo Dataset."
            )

        with TemporaryDirectory() as tmpdir:
            # Download the dataset from Zenodo
            zenodo = Zenodo(access_token=ZENODO_API_TOKEN)
            logger.info(f"Downloading {self.record_id} from Zenodo to {tmpdir}")
            file = zenodo.download_latest(
                self.record_id,
                name=self.name,
                force=True,
                parts=[str(tmpdir)],
            )
            if not self.root.exists():
                self.root.mkdir(parents=True, exist_ok=True)

            logger.info(f"Downloading {self.record_id} from Zenodo to {tmpdir}")
            self._move_data_to_root(file)


class ZenodoZipBase(ZenodoDataset):
    """Base class for Zenodo datasets that are zipped."""

    def __init__(
        self,
        root: str | Path,
        record_id: str,
        name: str,
        override: bool = False,
    ) -> None:
        super().__init__(root, record_id, name, override)

    def _move_data_to_root(self, file: Path) -> None:
        """Move the data to the root directory."""
        extract_archive(file, str(self.root))



from abc import abstractmethod
from logging import getLogger
from pathlib import Path

import kagglehub

from any_gold.utils.dataset import AnyDataset

logger = getLogger(__name__)


class KaggleDataset(AnyDataset):
    """Base class for Kaggle datasets.

    Kaggle is a platform for machine learning competitions and some datasets are made accessible for those competitions.
    It is accessible at https://huggingface.co/.

    This class specifies the basic way to download data from Kaggle.

    Attributes:
        root: The root directory where the dataset is stored.
        handle: The name of the dataset on Kaggle (same as the Kaggle dataset URL).
        override: If True, will override the existing dataset in the root directory. Default is False.
    """

    def __init__(
        self,
        root: str | Path,
        handle: str,
        override: bool = False,
    ) -> None:
        super().__init__(root=root, override=override)

        self.handle = handle

        self._setup()

    @abstractmethod
    def _setup(self) -> None:
        """download the data from Kaggle and initialize the elements of the dataset."""

    @abstractmethod
    def _move_data_to_root(self, kaggle_cache: Path) -> None:
        """Make the data available in the root directory.

        This method can be used to extract the data from an archive or to reorganise the data after downloading it.
        """

    def download(self) -> None:
        """Download the data from kaggle and store the dataset in the root folder."""
        logger.info(f"Downloading {self.handle} to kagglehub cache.")
        kaggle_cache = kagglehub.dataset_download(
            self.handle, force_download=self.override
        )
        self._move_data_to_root(Path(kaggle_cache))



from .utils.dataset import AnyRawDataset, AnyVisionSegmentationDataset
from .image.plantseg import PlantSeg
from .image.deepglobe import DeepGlobeRoadExtraction
from .image.kpi import KPITask1PatchLevel
from .image.mvtec_ad import MVTecADDataset

__all__ = (
    "AnyVisionSegmentationDataset",
    "AnyRawDataset",
    "PlantSeg",
    "DeepGlobeRoadExtraction",
    "KPITask1PatchLevel",
    "MVTecADDataset",
)



from pathlib import Path
from typing import Callable

from PIL import Image
import pandas as pd
import torch
from torchvision.tv_tensors import Image as TvImage, Mask as TvMask

from any_gold.utils.dataset import (
    AnyVisionSegmentationOutput,
    AnyVisionSegmentationDataset,
)
from any_gold.utils.zenodo import ZenodoZipBase


class PlantSegOutput(AnyVisionSegmentationOutput):
    """Output class for PlantSeg dataset.

    It extends the AnyVisionSegmentationOutput class to include plant species.
    The label will be the disease on the plant.
    """

    plant: str


class PlantSeg(AnyVisionSegmentationDataset, ZenodoZipBase):
    """PlantSeg Dataset from Zenodo.

    The PlantSeg dataset is introduced in
    [PlantSeg: A Large-Scale In-the-wild Dataset for Plant Disease Segmentation](https://arxiv.org/abs/2409.04038)

    The dataset is a collection of images and their corresponding segmentation masks for plant diseases (1 mask per image).
    The dataset is available in three versions, each with different images and masks.

    The dataset is downloaded from [Zenodo](https://zenodo.org/records/14935094)
    and its data will be downloaded and stored in the specified root directory.

    There are 3 different splits available: train, val, and test.

    Attributes:
        root: The root directory where the dataset is stored.
        version: The version of the dataset to use. Default is 3.
        split: The split of the dataset to use. Can be 'train', 'val', or 'test'. Default is 'train'.
        record_id: The record ID of the dataset on Zenodo.
        name: The name of the dataset on Zenodo.
        transform: A transform to apply to the images.
        target_transform: A transform to apply to the masks.
        transforms: A transform to apply to both images and masks.
        It cannot be set together with transform and target_transform.
        override: If True, will override the existing dataset in the root directory. Default is False.
        samples: A list of tuples containing the image path, plant name, and disease name.
    """

    _VERSIONS = {
        1: {
            "record_id": "13762907",
            "name": "plantseg.zip",
        },
        2: {
            "record_id": "13958858",
            "name": "plantsegv2.zip",
        },
        3: {
            "record_id": "14935094",
            "name": "plantsegv3.zip",
            "metadata": "Metadatav2.csv",
        },
    }

    def __init__(
        self,
        root: str | Path,
        version: int = 3,
        split: str = "train",
        transform: Callable | None = None,
        target_transform: Callable | None = None,
        transforms: Callable | None = None,
        override: bool = False,
    ) -> None:
        AnyVisionSegmentationDataset.__init__(
            self,
            root=root,
            transform=transform,
            target_transform=target_transform,
            transforms=transforms,
        )

        if version not in self._VERSIONS:
            raise ValueError(
                f"Version {version} is not available. Available versions are {list(self._VERSIONS.keys())}."
            )
        self.version = version

        if split not in ("train", "val", "test"):
            raise ValueError(
                f"Split {split} is not available. Available splits are ['train', 'val', 'test']."
            )
        self.split = split

        ZenodoZipBase.__init__(
            self,
            root=root,
            record_id=self._VERSIONS[version]["record_id"],
            name=self._VERSIONS[version]["name"],
            override=override,
        )

    def _setup(self) -> None:
        if self.override or not self.root.exists():
            self.download()

        self.samples: list[tuple[Path, str, str]] = []
        metadata = pd.read_csv(
            self.root
            / f"plantsegv{self.version}/{self._VERSIONS[self.version]['metadata']}"
        )
        for image_path in (
            self.root / f"plantsegv{self.version}/images/{self.split}"
        ).glob("*.jpg"):
            row = metadata[metadata["Name"] == image_path.name]
            if len(row) != 1:
                raise ValueError(
                    "Expected exactly one row per image in the PlantSeg metadata file."
                )

            self.samples.append(
                (image_path, row.iloc[0]["Plant"], row.iloc[0]["Disease"])
            )

    def __len__(self) -> int:
        return len(self.samples)

    def get_image_path(self, index: int) -> Path:
        """Get the path of an image."""
        return self.samples[index][0]

    def get_raw(self, index: int) -> PlantSegOutput:
        """Get the image and its corresponding mask together with the plant species, disease and index."""
        image_path, plant, disease = self.samples[index]
        mask_path = (
            image_path.parent.parent.parent
            / f"annotations/{self.split}/{image_path.stem}.png"
        )

        image = TvImage(Image.open(image_path).convert("RGB"), dtype=torch.uint8)
        mask = (
            TvMask(Image.open(mask_path).convert("L"), dtype=torch.uint8)
            if mask_path is not None
            else torch.zeros((1, *image.shape[-2:]), dtype=torch.uint8)
        )

        return PlantSegOutput(
            index=index,
            image=image,
            mask=mask,
            plant=plant,
            label=disease,
        )



from pathlib import Path
from typing import Callable

import torch
from torchvision.tv_tensors import Image as TvImage, Mask as TvMask

from any_gold.utils.dataset import (
    AnyVisionSegmentationDataset,
    AnyVisionSegmentationOutput,
)
from any_gold.utils.hugging_face import HuggingFaceDataset


class MVTecADOutput(AnyVisionSegmentationOutput):
    """Output class for MVTec Anomaly Detection dataset.

    It extends the AnyVisionSegmentationOutput class to include
    target (torch tensor indicating if the label is an anomaly or not).

    The label is the type of defect (`good` in absence of defect).
    """

    target: torch.Tensor


class MVTecADDataset(AnyVisionSegmentationDataset, HuggingFaceDataset):
    """MVTec Anomaly Detection Dataset.

    The Mvtec Anomaly Detection dataset is introduced in
    [The MVTec Anomaly Detection Dataset: A Comprehensive Real-World Dataset for Unsupervised Anomaly Detection](https://link.springer.com/content/pdf/10.1007/s11263-020-01400-4.pdf)

    The dataset is a collection of images and their corresponding segmentation masks for various
    manufacturing objects and type of defects.

    The dataset is downloaded from [Hugging Face](https://huggingface.co/datasets/TheoM55/mvtec_all_objects_split)
    and its data will be downloaded and stored in the specified root directory.

    For each category, there are two splits available: train and test.

    Args:
        root: The root directory where the dataset is stored.
        category: The category of the dataset (e.g., 'bottle', 'cable', 'capsule').
        split: The dataset split to use ('train' or 'test').
        path: The path of the dataset on Hugging Face.
        hf_split: it will be category.split.
        transform: A transform to apply to the images.
        target_transform: A transform to apply to the masks.
        transforms: A transform to apply to both images and masks.
        It cannot be set together with transform and target_transform.
        override: If True, will override the existing dataset in the root directory. Default is False.
        samples: the hugging face dataset in torch format.
    """

    _SPLITS = ("train", "test")
    _HUGGINGFACE_NAME = "TheoM55/mvtec_all_objects_split"

    def __init__(
        self,
        root: str | Path,
        category: str,
        split: str = "train",
        transform: Callable | None = None,
        target_transform: Callable | None = None,
        transforms: Callable | None = None,
        override: bool = False,
    ) -> None:
        self.category = category
        self.split = split
        if split not in self._SPLITS:
            raise ValueError(f"Split must be one of {self._SPLITS}, but got {split}.")

        HuggingFaceDataset.__init__(
            self,
            path=self._HUGGINGFACE_NAME,
            hf_split=f"{self.category}.{self.split}",
            override=override,
        )
        AnyVisionSegmentationDataset.__init__(
            self,
            root=root,
            transform=transform,
            target_transform=target_transform,
            transforms=transforms,
            override=override,
        )

    def __len__(self) -> int:
        return len(self.samples)

    def get_raw(self, index: int) -> MVTecADOutput:
        """
        Get the image and its corresponding mask together with the label,
        the anomaly detection target and the index.
        """
        sample = self.samples[index]

        # image_path and mask_path are already torch tensors
        image = TvImage(sample["image_path"])
        mask = (
            TvMask(sample["mask_path"])
            if sample["mask_path"] is not None
            else torch.zeros((1, *image.shape[-2:]), dtype=torch.uint8)
        )

        return MVTecADOutput(
            image=image,
            mask=mask,
            label=sample["defect"],
            target=sample["label"],
            index=index,
        )



from abc import abstractmethod
from collections import defaultdict
from pathlib import Path
from typing import Callable, TypedDict, Any

from torch.utils.data import Dataset, DataLoader
from torchvision.datasets import VisionDataset
from torchvision.tv_tensors import Image as TvImage, Mask as TvMask
from tqdm import tqdm

from any_gold.tools.image.connected_component import (
    extract_connected_components_from_binary_mask,
)
from any_gold.tools.image.stats import vision_segmentation_stats


class AnyOutput(TypedDict):
    """Base class for any output from a dataset.

    It specifies the output format for any dataset. All any gold dataset is expected to return an instance of this class.
    At least the index of the data is required, but it can also contain
    other information such as image, annotation, etc...
    """

    index: int


class AnyDataset(Dataset):
    """Base class for any dataset.

    Attributes:
        root: The root directory where the dataset is stored.
        override: If True, will override the existing dataset in the root directory.
    """

    def __init__(
        self,
        root: str | Path,
        override: bool = False,
    ) -> None:
        super().__init__()

        self.root = Path(root)
        self.override = override

    @abstractmethod
    def get_raw(self, index: int) -> AnyOutput:
        """Get the raw data for the given index."""

    @abstractmethod
    def __get_item__(self, index: int) -> AnyOutput:
        """Get the transformed data for the given index."""

    @abstractmethod
    def __len__(self) -> int:
        """Get the length of the dataset."""

    @abstractmethod
    def describe(self, batch_size: int = 1, num_workers: int = 0) -> dict[str, Any]:
        """Get a description of the dataset, including the number of samples and other relevant information."""


class AnyVisionSegmentationOutput(AnyOutput):
    """Base class for any vision segmentation output.

    It specifies the output format for any vision segmentation dataset.
    It extends the AnyOutput class to include image and mask attributes.
    """

    image: TvImage
    mask: TvMask
    label: str


class AnyVisionSegmentationDataset(VisionDataset, AnyDataset):
    """Base class for any vision dataset.

    The image and mask are expected to be in the torchvision format with a shape C, H, W for images and masks.

    Attributes:
        root: The root directory where the dataset is stored.
        transform: A transform to apply to the images.
        target_transform: A transform to apply to the masks.
        transforms: A transform to apply to both images and masks.
        It cannot be set together with transform and target_transform.
        override: If True, will override the existing dataset in the root directory. Default is False.
    """

    def __init__(
        self,
        root: str | Path,
        transform: Callable | None = None,
        target_transform: Callable | None = None,
        transforms: Callable | None = None,
        override: bool = False,
    ) -> None:
        AnyDataset.__init__(self, root, override)

        VisionDataset.__init__(
            self,
            root=root,
            transform=transform,
            target_transform=target_transform,
            transforms=transforms,
        )

    @abstractmethod
    def get_raw(self, index: int) -> AnyVisionSegmentationOutput:
        """Get the raw data for the given index."""

    def describe(self, batch_size: int = 1, num_workers: int = 0) -> dict[str, Any]:
        """Make a description of the dataset.

        Args:
            batch_size: The batch size to use for the DataLoader.
            num_workers: The number of workers to use for the DataLoader.

        Returns:
            A dictionary containing the description of the dataset, including:
            - name: The name of the dataset class.
            - sample count: The number of samples in the dataset.
            - shapes: A dictionary with the shape as key and the corresponding number of images in the dataset.
            - areas: A dictionary with min, max, mean, and total area of connected components.
            - object count: A dictionary with min, max, mean, and total object count.
            - per label statistics if applicable.
        """
        dataloader = DataLoader(
            self,
            batch_size=batch_size,
            num_workers=num_workers,
            collate_fn=lambda x: x,  # some datasets might contain different sizes of images and masks
        )
        global_shape_count: dict[tuple[int, ...], int] = defaultdict(int)
        global_areas = []
        global_object_count = []
        shape_count_per_label: dict[str, dict[tuple[int, ...], int]] = defaultdict(
            lambda: defaultdict(int)
        )
        areas_per_label = defaultdict(list)
        object_count_per_label = defaultdict(list)

        for batch in tqdm(
            dataloader,
            desc=f"Describing dataset {self.__class__.__name__}",
            unit="batch",
        ):
            for sample in batch:
                label = sample["label"]

                shape = tuple(sample["image"].shape)
                global_shape_count[shape] += 1
                shape_count_per_label[label][shape] += 1

                connected_components = extract_connected_components_from_binary_mask(
                    sample["mask"].permute(1, 2, 0).numpy(), min_area=1
                )

                if not connected_components:
                    continue

                areas = [cc.area for cc in connected_components]
                global_areas.extend(areas)
                areas_per_label[label].extend(areas)

                object_count = len(connected_components)
                global_object_count.append(object_count)
                object_count_per_label[label].append(object_count)

        global_description = vision_segmentation_stats(
            shape_counts=global_shape_count,
            areas=global_areas,
            object_counts=global_object_count,
        )
        label_description = {
            label: vision_segmentation_stats(
                shape_counts=shape_count_per_label[label],
                areas=areas,
                object_counts=object_count_per_label[label],
            )
            for label, areas in areas_per_label.items()
        }

        return (
            {"name": self.__class__.__name__, "sample count": len(self)}
            | global_description
            | label_description
        )

    def __getitem__(self, index: int) -> AnyVisionSegmentationOutput:
        """Get the transformed image and its corresponding information."""
        output = self.get_raw(index)

        image = output["image"]
        mask = output["mask"]

        if self.transform:
            output["image"] = self.transform(image)
        if self.target_transform:
            output["mask"] = self.target_transform(mask)
        if self.transforms:
            output["image"], output["mask"] = self.transforms(image, mask)

        return output


class AnyRawDataset:
    """Wrapper allowing to load raw data from a dataset.

    Attributes:
        dataset: The dataset to wrap.
    """

    def __init__(
        self,
        dataset: AnyDataset,
    ) -> None:
        self._dataset = dataset

    def __getitem__(self, index: int) -> AnyOutput:
        """Get the raw data of the dataset for the given index."""
        return self._dataset.get_raw(index)

    def __len__(self) -> int:
        """Get the length of the dataset."""
        return len(self._dataset)



import shutil
from pathlib import Path
from typing import Callable

from PIL import Image
import torch
from torchvision.tv_tensors import Image as TvImage, Mask as TvMask

from any_gold.utils.dataset import (
    AnyVisionSegmentationDataset,
    AnyVisionSegmentationOutput,
)
from any_gold.utils.kaggle import KaggleDataset


class DeepGlobeRoadExtractionOutput(AnyVisionSegmentationOutput):
    """Output class for DeepGlobe Road Extraction dataset.

    The label will always be `road`.
    """

    pass


class DeepGlobeRoadExtraction(AnyVisionSegmentationDataset, KaggleDataset):
    """Deepglobe road extraction dataset from kaggle.

    The DeepGlobe road extraction dataset is introduced in
    [DeepGlobe 2018: A Challenge to Parse the Earth Through Satellite Images](https://arxiv.org/pdf/1805.06561)

    The dataset is a collection of satellite images with road to be extracted from.
    Only the training set is integrating road location (1 mask per image)

    The dataset is downloaded from [Kaggle](https://www.kaggle.com/datasets/balraj98/deepglobe-road-extraction-dataset)
    and its data will be downloaded and stored in the specified root directory.

    There are 3 different splits available: train, val, and test.

    Attributes:
        root: The root directory where the dataset is stored.
        split: The split of the dataset to use. Can be 'train', 'val', or 'test'. Default is 'train'.
        handle: The name of the dataset on Kaggle (same as _HANDLE).
        transform: A transform to apply to the images.
        target_transform: A transform to apply to the masks.
        transforms: A transform to apply to both images and masks.
        It cannot be set together with transform and target_transform.
        override: If True, will override the existing dataset in the root directory. Default is False.
        samples: A list of file paths to the satellite images in the specified split.
    """

    _HANDLE = "balraj98/deepglobe-road-extraction-dataset/versions/2"
    _SPLITS = {
        "train": "train",
        "val": "valid",
        "test": "test",
    }

    def __init__(
        self,
        root: str | Path,
        split: str = "train",
        transform: Callable | None = None,
        target_transform: Callable | None = None,
        transforms: Callable | None = None,
        override: bool = False,
    ) -> None:
        if split not in self._SPLITS:
            raise ValueError(f"Split must be one of {self._SPLITS}, but got {split}.")

        self.split = split

        AnyVisionSegmentationDataset.__init__(
            self,
            root=root,
            transform=transform,
            target_transform=target_transform,
            transforms=transforms,
        )
        KaggleDataset.__init__(
            self,
            root=root,
            handle=self._HANDLE,
            override=override,
        )

    def _move_data_to_root(self, kaggle_cache: Path) -> None:
        self.root.mkdir(parents=True, exist_ok=True)

        target = self.root / self.split
        source = kaggle_cache / self._SPLITS[self.split]

        if target.exists():
            shutil.rmtree(target)

        if not source.exists():
            raise FileNotFoundError(
                f"Source directory {source} does not exist. "
                f"Please use override=True to download the dataset again."
            )

        shutil.move(source, target)

    def _setup(self) -> None:
        root = self.root / self.split
        if self.override or not root.exists():
            self.download()

        self.samples = [image_path for image_path in root.glob("*_sat.jpg")]

    def __len__(self) -> int:
        return len(self.samples)

    def get_image_path(self, index: int) -> Path:
        """Get the path of an image."""
        return self.samples[index]

    def get_raw(self, index: int) -> DeepGlobeRoadExtractionOutput:
        """Get an image and its corresponding mask together with the index.

        If the split is not 'train', the mask will be None.
        """
        image_path = self.samples[index]
        mask_path = image_path.parent / f"{image_path.stem[:-3]}mask.png"

        image = TvImage(Image.open(image_path).convert("RGB"), dtype=torch.uint8)
        mask = (
            TvMask(Image.open(mask_path).convert("L"), dtype=torch.uint8)
            if mask_path is not None
            else torch.zeros((1, *image.shape[-2:]), dtype=torch.uint8)
        )

        return DeepGlobeRoadExtractionOutput(
            image=image, mask=mask, index=index, label="road"
        )



import os
from abc import abstractmethod
from pathlib import Path
from tempfile import TemporaryDirectory
from logging import getLogger

from torchvision.datasets.utils import extract_archive

import synapseclient
import synapseutils

from any_gold.utils.dataset import AnyDataset

logger = getLogger(__name__)


class SynapseDataset(AnyDataset):
    """Base class for Synapse datasets.

    Synapse is an open repository for biomedical research data and software.
    It is accessible at https://www.synapse.org/.

    This class specifies the basic way to download data from Synapse. Each inherited class must implement the
    `_move_data_to_root` and `_setup` methods to download the data from Synapse.

    The SYNAPSE_API_TOKEN environment variable must be set to access Synapse datasets.

    Attributes:
        root: The root directory where the dataset is stored.
        override: If True, will override the existing dataset in the root directory. Default is False.
        entity: The Synapse entity ID of the dataset.
    """

    def __init__(
        self,
        root: str | Path,
        entity: str,
        override: bool = False,
    ) -> None:
        super().__init__(
            root=root,
            override=override,
        )

        self.entity = entity

        self._setup()

    @abstractmethod
    def _setup(self) -> None:
        """download the data from Synapse and initialize the elements of the dataset."""

    @abstractmethod
    def _move_data_to_root(self, files: list[synapseclient.entity.File]) -> None:
        """Make the data available in the root directory.

        This method can be used to extract the data from an archive or to reorganise the data after downloading it.
        """

    def download(self) -> None:
        """Download the data from Synapse and store the dataset in the root folder."""
        SYNAPSE_API_TOKEN = os.environ.get("SYNAPSE_API_TOKEN")
        if SYNAPSE_API_TOKEN is None:
            raise ValueError(
                "Please set the SYNAPSE_API_TOKEN environment variable to access Synapse Dataset."
            )

        with TemporaryDirectory() as tmpdir:
            syn = synapseclient.Synapse()
            syn.login(authToken=SYNAPSE_API_TOKEN)

            logger.info(f"Downloading {self.entity} from Synapse to {tmpdir}")
            files = synapseutils.syncFromSynapse(syn, self.entity, str(tmpdir))

            logger.info(f"Moving files to {self.root}")
            self._move_data_to_root(files)


class SynapseZipBase(SynapseDataset):
    """Base class for Synapse datasets that are zipped."""

    def __init__(
        self,
        root: str | Path,
        entity: str,
        override: bool = False,
    ) -> None:
        super().__init__(root, entity, override)

    def _move_data_to_root(self, files: list[synapseclient.entity.File]) -> None:
        """Move the data to the root directory."""
        for file in files:
            extract_archive(file.path, str(self.root))
